from __future__ import unicode_literals

import logging
import re

from django.contrib.contenttypes.models import ContentType
from django.db import connections, router

from .. import signals
from ..compat import get_remote_field


def create_tenant_schema(tenant, using=None):
    """
    CREATE the tables associated with a tenant's models.
    """
    logger = logging.getLogger('tenancy.management.create_tenant_schema')
    tenant_class = tenant.__class__
    using = using or router.db_for_write(tenant_class, instance=tenant)
    connection = connections[using]
    quote_name = connection.ops.quote_name

    tenant_class._default_manager._add_to_cache(tenant)

    signals.pre_schema_creation.send(
        sender=tenant_class, tenant=tenant, using=using
    )

    if connection.vendor == 'postgresql':
        schema = tenant.db_schema
        quoted_schema = quote_name(schema)
        from ..settings import SCHEMA_AUTHORIZATION
        if SCHEMA_AUTHORIZATION:
            create_schema = "CREATE SCHEMA AUTHORIZATION %s"
        else:
            create_schema = "CREATE SCHEMA %s"
        logger.info("Creating schema %s ..." % schema)
        connection.cursor().execute(create_schema % quoted_schema)

    signals.post_schema_creation.send(
        sender=tenant_class, tenant=tenant, using=using
    )

    signals.pre_models_creation.send(
        sender=tenant_class, tenant=tenant, using=using
    )

    with connection.schema_editor() as editor:
        for model in tenant.models:
            # Has the side effect of creating the required `ContentType`.
            ContentType.objects.get_for_model(model)
            # Avoid further processing we're dealing with an unmanaged model or
            # one proxying another.
            opts = model._meta
            if not opts.managed or opts.proxy or opts.auto_created:
                continue
            if not router.allow_migrate(connection.alias, model):
                continue
            logger.debug(
                "Processing %s.%s model" % (opts.app_label, opts.object_name)
            )
            if connection.vendor == 'postgresql':
                table_name = "%s.%s" % (
                    schema, model._for_tenant_model._meta.db_table
                )
            else:
                table_name = opts.db_table
            logger.info("Creating table %s ..." % table_name)
            for m2m in opts.many_to_many:
                through_opts = get_remote_field(m2m).through._meta
                if through_opts.auto_created:
                    logger.info("Creating table %s ..." % through_opts.db_table)
            editor.create_model(model)
            if connection.vendor == 'postgresql' and SCHEMA_AUTHORIZATION:
                quoted_tables = [quote_name(opts.db_table)] + [
                    quote_name(get_remote_field(m2m).through._meta.db_table)
                    for m2m in opts.many_to_many if through_opts.auto_created
                ]
                editor.deferred_sql.extend(
                    "ALTER TABLE %s OWNER TO %s" % (
                        quoted_table, quoted_schema
                    ) for quoted_table in quoted_tables
                )
        if connection.vendor == 'postgresql':
            altered_statements = []
            # Our "db_table" hack to allow specifying a schema interferes with
            # index and constraint creation.
            create_index_re = re.compile('CREATE INDEX %s' % re.escape("%s." % quoted_schema))
            add_constraint_re = re.compile('ADD CONSTRAINT "([^\s]+)" ')
            for statement in editor.deferred_sql:
                if statement.startswith('CREATE INDEX'):
                    statement = create_index_re.sub('CREATE INDEX ', statement)
                elif 'ADD CONSTRAINT' in statement:
                    statement = add_constraint_re.sub(
                        lambda match: 'ADD CONSTRAINT "%s" ' % match.group(1).replace('"', '_'),
                        statement
                    )
                altered_statements.append(statement)
            editor.deferred_sql = altered_statements

    signals.post_models_creation.send(
        sender=tenant_class, tenant=tenant, using=using
    )


def drop_tenant_schema(tenant, using=None):
    """
    DROP the tables associated with a tenant's models.
    """
    tenant_class = tenant.__class__
    using = using or router.db_for_write(tenant_class, instance=tenant)
    connection = connections[using]
    quote_name = connection.ops.quote_name

    signals.pre_schema_deletion.send(
        sender=tenant_class, tenant=tenant, using=using
    )

    ContentType.objects.filter(
        pk__in=[
            ct.pk for ct in ContentType.objects.get_for_models(
                *tenant.models, for_concrete_models=False
            ).values()
        ]
    ).delete()

    if connection.vendor == 'postgresql':
        connection.cursor().execute(
            "DROP SCHEMA %s CASCADE" % quote_name(tenant.db_schema)
        )
    else:
        with connection.schema_editor() as editor:
            for model in tenant.models:
                opts = model._meta
                if not opts.managed or opts.proxy or opts.auto_created:
                    continue
                if not router.allow_migrate(connection.alias, model):
                    continue
                editor.delete_model(model)

    tenant_class._default_manager._remove_from_cache(tenant)
    ContentType.objects.clear_cache()

    signals.post_schema_deletion.send(
        sender=tenant_class, tenant=tenant, using=using
    )



from __future__ import unicode_literals

import threading

from django.db import models

try:
    from django.db.models.query import ModelIterable
except ImportError:
    # TODO: Remove when dropping support for Django 1.8.
    class TenantQueryset(models.QuerySet):
        def iterator(self):
            iterator = super(TenantQueryset, self).iterator()
            add_to_cache = self.model._default_manager._add_to_cache
            for tenant in iterator:
                yield add_to_cache(tenant)
else:
    class TenantIterable(ModelIterable):
        def __iter__(self):
            iterator = super(TenantIterable, self).__iter__()
            add_to_cache = self.queryset.model._default_manager._add_to_cache
            for tenant in iterator:
                yield add_to_cache(tenant)

    class TenantQueryset(models.QuerySet):
        def __init__(self, *args, **kwargs):
            super(TenantQueryset, self).__init__(*args, **kwargs)
            self._iterable_class = TenantIterable


class AbstractTenantManager(models.Manager.from_queryset(TenantQueryset)):
    __cache = threading.local()

    @property
    def _tenants(self):
        try:
            return self.__cache.tenants
        except AttributeError:
            tenants = {}
            setattr(self.__cache, 'tenants', tenants)
            return tenants

    def clear_cache(self):
        for tenant in list(self._tenants.values()):
            self._remove_from_cache(tenant)

    def should_cache(self, tenant):
        """Return whether or not a tenant instance should be cached."""
        # TODO: Remove when dropping support for Django 1.9 and simply return True.
        return not getattr(tenant, '_deferred', False)

    def _get_from_cache(self, *natural_key):
        return self._tenants[natural_key]

    def _add_to_cache(self, tenant):
        if self.should_cache(tenant):
            key = tenant.natural_key()
            try:
                return self._get_from_cache(*key)
            except KeyError:
                self._tenants[key] = tenant
        return tenant

    def _remove_from_cache(self, tenant):
        key = tenant.natural_key()
        delattr(tenant, 'models')
        tenant.__class__.models.tenant_models.pop(key, None)
        return self._tenants.pop(key)

    def _get_by_natural_key(self, *natural_key):
        raise NotImplementedError

    def get_by_natural_key(self, *natural_key):
        try:
            tenant = self._get_from_cache(*natural_key)
        except KeyError:
            tenant = self._add_to_cache(self._get_by_natural_key(*natural_key))
        return tenant


class TenantManager(AbstractTenantManager):
    def _get_by_natural_key(self, name):
        return self.get(name=name)


class TenantModelManagerDescriptor(object):
    """
    This class provides a better error message when you try to access a
    manager on an tenant model.
    """
    def __init__(self, model):
        self.model = model

    def __get__(self, instance, owner):
        raise AttributeError(
            "Manager isn't available; %s is tenant specific" % (
                self.model._meta.object_name,
            )
        )



from __future__ import unicode_literals

import warnings

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Manager
from django.db.models.query import QuerySet
from django.forms.models import (
    BaseInlineFormSet, BaseModelFormSet, ModelForm, modelform_factory,
)
from django.utils.functional import cached_property
from django.views.generic.edit import ModelFormMixin

from . import get_tenant_model
from .forms import (
    tenant_inlineformset_factory, tenant_modelform_factory,
    tenant_modelformset_factory,
)
from .models import TenantModelBase, TenantSpecificModel


class TenantMixin(object):
    """
    View mixin that retrieve the current tenant from the request. This could
    have been set from a middleware base on a domain name for example.
    """
    tenant_attr_name = get_tenant_model().ATTR_NAME

    def get_tenant(self):
        return getattr(self.request, self.tenant_attr_name)


setattr(
    TenantMixin, get_tenant_model().ATTR_NAME,
    cached_property(lambda self: self.get_tenant())
)


class TenantObjectMixin(TenantMixin):
    """
    View mixin that returns the correct queryset for the specified model based
    on the retrieved tenant.
    """
    model = None
    context_object_name = None

    def get_model(self):
        if self.model:
            if not isinstance(self.model, TenantModelBase):
                msg = "%s.model is not an instance of TenantModelBase."
                raise ImproperlyConfigured(msg % self.__class__.__name__)
            return self.model
        raise ImproperlyConfigured(
            "%s is missing a model." % self.__class__.__name__
        )

    def get_tenant_model(self):
        tenant = self.get_tenant()
        return tenant.models[self.get_model()]

    def get_queryset(self):
        return self.get_tenant_model()._default_manager.all()

    def get_template_names(self):
        try:
            names = super(TenantObjectMixin, self).get_template_names()
        except (AttributeError, ImproperlyConfigured):
            names = []

        model = self.get_model()
        names.append("%s/%s%s.html" % (
            model._meta.app_label,
            model._meta.model_name,
            getattr(self, 'template_name_suffix', ''),
        ))

        return names

    def get_context_object_name(self, obj):
        if self.context_object_name:
            return self.context_object_name
        elif (isinstance(obj, (Manager, QuerySet)) and
              issubclass(obj.model, TenantSpecificModel)):
            return "%s_list" % obj.model._for_tenant_model._meta.model_name
        elif isinstance(obj, TenantSpecificModel):
            return obj._for_tenant_model._meta.model_name


class TenantModelFormMixin(TenantObjectMixin, ModelFormMixin):
    def get_form_class(self):
        """
        Provide a model form class based on tenant specific model.
        """
        form_class = self.form_class
        fields = self.fields
        if form_class:
            if issubclass(form_class, ModelForm):
                model = form_class._meta.model
                factory = tenant_modelform_factory
            elif issubclass(form_class, BaseModelFormSet):
                model = form_class.model
                if issubclass(form_class, BaseInlineFormSet):
                    factory = tenant_inlineformset_factory
                else:
                    factory = tenant_modelformset_factory
            else:
                raise ImproperlyConfigured(
                    "%s.form_class must be a subclass of `ModelForm` or "
                    "`BaseModelFormSet`." % self.__class__.__name__
                )
            if model:
                if not isinstance(model, TenantModelBase):
                    raise ImproperlyConfigured(
                        "%s.form_class' model is not an instance of "
                        "TenantModelBase." % self.__class__.__name__
                    )
                return factory(self.get_tenant(), form_class)
        else:
            form_class = ModelForm
            if fields is None:
                warnings.warn(
                    "Using TenantModelFormMixin (base class of %s) without "
                    "the 'fields' attribute is deprecated." % self.__class__.__name__,
                    DeprecationWarning,
                )
        return modelform_factory(
            self.get_tenant_model(), form_class, fields=fields
        )


class TenantWizardMixin(TenantMixin):
    def get_form(self, step=None, data=None, files=None):
        if step is None:
            step = self.steps.current
        kwargs = self.get_form_kwargs(step)
        form_class = self.form_list[step]
        if (issubclass(form_class, ModelForm) and
                isinstance(form_class._meta.model, TenantModelBase)):
            kwargs.setdefault('instance', self.get_form_instance(step))
            form_class = tenant_modelform_factory(self.get_tenant(), form_class)
        elif (issubclass(form_class, BaseModelFormSet) and
              isinstance(form_class.model, TenantModelBase)):
            kwargs.setdefault('queryset', self.get_form_instance(step))
            tenant = self.get_tenant()
            if isinstance(form_class, BaseInlineFormSet):
                form_class = tenant_inlineformset_factory(tenant, form_class)
            else:
                form_class = tenant_modelformset_factory(tenant, form_class)
        else:
            return super(TenantWizardMixin, self).get_form(step, data, files)
        kwargs.update({
            'data': data,
            'files': files,
            'prefix': self.get_form_prefix(step, form_class),
            'initial': self.get_form_initial(step),
        })
        return form_class(**kwargs)



from __future__ import unicode_literals

import warnings

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured

from .. import get_tenant_model
from ..models import TenantModelBase


class TenantUserBackend(object):
    def __init__(self):
        self.user_model = get_user_model()
        self.tenant_model = get_tenant_model()
        if not isinstance(self.user_model, TenantModelBase):
            raise ImproperlyConfigured(
                "The `tenancy.auth.backends.TenantUserBackend` "
                "authentification backend can only be used with a custom "
                "tenant user model."
            )

    def get_tenant_user_model(self, tenant=None):
        if tenant is None:
            tenant = self.tenant_model.get_global()
            if tenant is None:
                return None
        return self.user_model.for_tenant(tenant)

    def get_tenant_user_queryset(self, tenant):
        return self.get_tenant_user_model(tenant)._default_manager

    def authenticate(self, request=None, username=None, password=None, tenant=None, **kwargs):
        if request is not None:
            tenant = getattr(request, self.tenant_model.ATTR_NAME, None)
        if tenant is None:
            tenant = kwargs.get(self.tenant_model.ATTR_NAME)
        tenant_user_model = self.get_tenant_user_model(tenant)
        if tenant_user_model is None:
            return None
        users = self.get_tenant_user_queryset(tenant)
        username_field = tenant_user_model.USERNAME_FIELD
        if username is None:
            username = kwargs.get(username_field)
        try:
            user = users.get(**{username_field: username})
        except tenant_user_model.DoesNotExist:
            return None
        if user.check_password(password):
            return user

    def get_user(self, pk):
        tenant_user_model = self.get_tenant_user_model()
        if tenant_user_model is None:
            return None
        tenant = getattr(tenant_user_model, self.tenant_model.ATTR_NAME)
        users = self.get_tenant_user_queryset(tenant)
        try:
            return users.get(pk=pk)
        except tenant_user_model.DoesNotExist:
            return None


class CustomTenantUserBackend(TenantUserBackend):
    def __init__(self):
        warnings.warn('CustomTenantUserBackend is deprecated in favor or TenantUserBackend.', DeprecationWarning)
        super(CustomTenantUserBackend, self).__init__()



from __future__ import unicode_literals

from django.core.signals import Signal
from django.db.models.signals import class_prepared

from .utils import get_model


def lazy_class_prepared(app_label, object_name, callback):
    """
    Lazily execute a callback upon model class preparation.
    """
    model = get_model(app_label, object_name.lower())
    if model:
        callback(model)
    else:
        def receiver(sender, **kwargs):
            opts = sender._meta
            if (opts.app_label == app_label and
                    opts.object_name == object_name):
                class_prepared.disconnect(receiver)
                callback(sender)
        class_prepared.connect(receiver, weak=False)


pre_schema_creation = Signal(providing_args=['tenant', 'using'])
post_schema_creation = Signal(providing_args=['tenant', 'using'])

pre_models_creation = Signal(providing_args=['tenant', 'using'])
post_models_creation = Signal(providing_args=['tenant', 'using'])

pre_schema_deletion = Signal(providing_args=['tenant', 'using'])
post_schema_deletion = Signal(providing_args=['tenant', 'using'])



from __future__ import unicode_literals

from django.utils.version import get_version

VERSION = (0, 3, 1, 'alpha', 4)

__version__ = get_version(VERSION)

default_app_config = 'tenancy.apps.TenancyConfig'


def get_tenant_model():
    from django.core.exceptions import ImproperlyConfigured
    from .models import AbstractTenant
    from .utils import get_model
    from .settings import TENANT_MODEL

    app_label, object_name = TENANT_MODEL.split('.')
    model_name = object_name.lower()
    tenant_model = get_model(app_label, model_name)
    if tenant_model is None:
        raise ImproperlyConfigured(
            "TENANCY_TENANT_MODEL refers to model '%s.%s' that has not "
            "been installed" % (app_label, object_name)
        )
    elif not issubclass(tenant_model, AbstractTenant):
        raise ImproperlyConfigured(
            "TENANCY_TENANT_MODEL refers to models '%s.%s' which is not a "
            "subclass of 'tenancy.AbstractTenant'" % (app_label, object_name))
    return tenant_model



from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured

from .models import TenantModelBase


def _get_tenant_model(tenant, model):
    if not isinstance(model, TenantModelBase):
        raise ImproperlyConfigured(
            "%s must be an instance of TenantModelBase" % model.__name__
        )
    return tenant.models[model]


def tenant_modelform_factory(tenant, form):
    tenant_model = _get_tenant_model(tenant, form._meta.model)

    name = str("%s_%s" % (form.__name__, tenant.model_name_prefix))

    attrs = {'model': tenant_model}
    parent = (object,)
    if hasattr(form, 'Meta'):
        parent = (form.Meta, object)
    Meta = type(str('Meta'), parent, attrs)

    return type(name, (form,), {'Meta': Meta})


def tenant_modelformset_factory(tenant, formset):
    tenant_model = _get_tenant_model(tenant, formset.model)
    tenant_modelform = tenant_modelform_factory(tenant, formset.form)

    name = str("%s_%s" % (formset.__name__, tenant.model_name_prefix))
    attrs = {'model': tenant_model, 'form': tenant_modelform}

    return type(name, (formset,), attrs)


def tenant_inlineformset_factory(tenant, formset):
    tenant_modelformset = tenant_modelformset_factory(tenant, formset)
    fk = tenant_modelformset.model._meta.get_field(formset.fk.name)
    tenant_modelformset.fk = fk
    return tenant_modelformset



from __future__ import unicode_literals

import copy
import logging
from abc import ABCMeta
from collections import OrderedDict
from contextlib import contextmanager

import django
from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db import connection, models
from django.db.models.base import ModelBase, subclass_exception
from django.db.models.deletion import DO_NOTHING
from django.db.models.fields import Field
from django.dispatch.dispatcher import receiver
from django.utils.deconstruct import deconstructible
from django.utils.six import itervalues, string_types, with_metaclass
from django.utils.six.moves import copyreg

from . import get_tenant_model, settings
from .compat import (
    get_private_fields, get_remote_field, get_remote_field_model,
    lazy_related_operation, set_remote_field_model,
)
from .management import create_tenant_schema, drop_tenant_schema
from .managers import (
    AbstractTenantManager, TenantManager, TenantModelManagerDescriptor,
)
from .signals import lazy_class_prepared
from .utils import (
    clear_cached_properties, clear_opts_related_cache, disconnect_signals,
    get_model, receivers_for_model, remove_from_app_cache,
)


class TenantModels(object):
    __slots__ = ['references']

    def __init__(self, tenant):
        self.references = OrderedDict((
            (reference, reference.for_tenant(tenant))
            for reference in TenantModelBase.references
        ))

    def __getitem__(self, key):
        return self.references[key]

    def __iter__(self, **kwargs):
        return itervalues(self.references, **kwargs)


class TenantModelsDescriptor(object):
    def contribute_to_class(self, cls, name):
        self.name = name
        self.tenant_models = {}
        setattr(cls, name, self)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        try:
            return instance.__dict__[self.name]
        except KeyError:
            pass

        tenant_key = instance.natural_key()
        try:
            models = self.tenant_models[tenant_key]
        except KeyError:
            models = self.tenant_models[tenant_key] = TenantModels(instance)
            self.__set__(instance, models)
        return models

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value

    def __delete__(self, instance):
        try:
            # Use the instance assigned values if available.
            models = instance.__dict__[self.name]
        except KeyError:
            tenant_key = instance.natural_key()
            try:
                models = self.tenant_models[tenant_key]
            except KeyError:
                return
        for model in models:
            model.destroy()


class AbstractTenant(models.Model):
    ATTR_NAME = 'tenant'

    objects = AbstractTenantManager()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        created = not self.pk
        save = super(AbstractTenant, self).save(*args, **kwargs)
        if created:
            create_tenant_schema(self)
        return save

    def delete(self, *args, **kwargs):
        delete = super(AbstractTenant, self).delete(*args, **kwargs)
        drop_tenant_schema(self)
        return delete

    def natural_key(self):
        raise NotImplementedError

    models = TenantModelsDescriptor()

    @contextmanager
    def as_global(self):
        """
        Expose this tenant as thread local object. This is required by parts
        of django relying on global states such as authentification backends.
        """
        setattr(connection, self.ATTR_NAME, self)
        try:
            yield
        finally:
            delattr(connection, self.ATTR_NAME)

    @classmethod
    def get_global(cls):
        return getattr(connection, cls.ATTR_NAME, None)

    @property
    def model_name_prefix(self):
        return "Tenant_%s" % '_'.join(self.natural_key())

    @property
    def db_schema(self):
        return "tenant_%s" % '_'.join(self.natural_key())


class Tenant(AbstractTenant):
    name = models.CharField(unique=True, max_length=20)

    objects = TenantManager()

    class Meta:
        swappable = 'TENANCY_TENANT_MODEL'

    def natural_key(self):
        return (self.name,)


@deconstructible
class Managed(object):
    """
    Sentinel object used to detect tenant managed models.
    """

    def __init__(self, tenant_model):
        self.tenant_model = tenant_model

    def __bool__(self):
        # Evaluates to False in order to prevent Django from managing the model.
        return False

    # Remove when dropping support for Python 2.7
    __nonzero__ = bool

    def __eq__(self, other):
        return isinstance(other, Managed) and other.tenant_model == self.tenant_model


def meta(Meta=None, **opts):
    """
    Create a class with specified opts as attributes to be used as model
    definition options.
    """
    if Meta:
        opts = dict(Meta.__dict__, **opts)
    return type(str('Meta'), (), opts)


def db_schema_table(tenant, db_table):
    if connection.vendor == 'postgresql':
        # See https://code.djangoproject.com/ticket/6148#comment:47
        return '%s\".\"%s' % (tenant.db_schema, db_table)
    else:
        return "%s_%s" % (tenant.db_schema, db_table)


class Reference(object):
    __slots__ = ['model', 'bases', 'Meta', 'related_names']

    def __init__(self, model, Meta, related_names=None):
        self.model = model
        self.Meta = Meta
        self.related_names = related_names

    def object_name_for_tenant(self, tenant):
        return "%s_%s" % (
            tenant.model_name_prefix,
            self.model._meta.object_name
        )

    def for_tenant(self, tenant):
        app_label = self.model._meta.app_label
        object_name = self.object_name_for_tenant(tenant)
        return "%s.%s" % (app_label, object_name)


class TenantSpecificModel(with_metaclass(ABCMeta)):
    @classmethod
    def __subclasshook__(cls, subclass):
        if isinstance(subclass, TenantModelBase):
            try:
                tenant_model = get_tenant_model()
            except ImproperlyConfigured:
                # If the tenant model is not configured yet we can assume
                # no specific models have been defined so far.
                return False
            tenant = getattr(subclass, tenant_model.ATTR_NAME, None)
            return isinstance(tenant, tenant_model)
        return NotImplemented


class TenantDescriptor(object):
    __slots__ = ['natural_key']

    def __init__(self, tenant):
        self.natural_key = tenant.natural_key()

    def __get__(self, model, owner):
        tenant_model = get_tenant_model()
        return tenant_model._default_manager.get_by_natural_key(*self.natural_key)


class TenantApps(object):
    def __init__(self, tenant, apps):
        self.apps = apps
        self.natural_key = tenant.natural_key()

    def get_models(self, *args, **kwargs):
        models = self.apps.get_models(*args, **kwargs)
        return [
            model for model in models
            if not issubclass(model, TenantSpecificModel) or model._meta.apps.natural_key == self.natural_key
        ]

    def __getattr__(self, name):
        return getattr(self.apps, name)


class TenantModelBase(ModelBase):
    reference = Reference
    references = OrderedDict()
    tenant_model_class = None
    exceptions = ('DoesNotExist', 'MultipleObjectsReturned')

    def __new__(cls, name, bases, attrs):
        super_new = super(TenantModelBase, cls).__new__

        Meta = attrs.setdefault('Meta', meta())
        if (getattr(Meta, 'abstract', False) or
                any(issubclass(base, TenantSpecificModel) for base in bases)):
            # Abstract model definition and ones subclassing tenant specific
            # ones shouldn't get any special treatment.
            model = super_new(cls, name, bases, attrs)
            if not cls.tenant_model_class:
                cls.tenant_model_class = model
        else:
            # Store managers to replace them with a descriptor specifying they
            # can't be accessed this way.
            managers = set(
                name for name, attr in attrs.items()
                if isinstance(attr, models.Manager)
            )
            # There's always a default manager named `objects`.
            managers.add('objects')
            if getattr(Meta, 'proxy', False):
                model = super_new(
                    cls, name, bases,
                    dict(attrs, meta=meta(Meta, managed=Managed(settings.TENANT_MODEL)))
                )
                cls.references[model] = cls.reference(model, Meta)
            else:
                # Extract field related names prior to adding them to the model
                # in order to validate them later on.
                related_names = {}
                for attr_name, attr in attrs.items():
                    if isinstance(attr, Field):
                        remote_field = get_remote_field(attr)
                        if remote_field:
                            related_names[attr.name or attr_name] = remote_field.related_name
                for base in bases:
                    if isinstance(base, ModelBase) and base._meta.abstract:
                        for field in base._meta.local_fields:
                            remote_field = get_remote_field(field)
                            if remote_field:
                                related_names[field.name] = remote_field.related_name
                        for m2m in base._meta.local_many_to_many:
                            related_names[m2m.name] = get_remote_field(m2m).related_name
                model = super_new(
                    cls, name, bases,
                    dict(attrs, Meta=meta(Meta, managed=Managed(settings.TENANT_MODEL)))
                )
                cls.references[model] = cls.reference(model, Meta, related_names)
                opts = model._meta
                # Validate related name of related fields.
                for field in (opts.local_fields + get_private_fields(opts)):
                    remote_field = get_remote_field(field)
                    if remote_field:
                        cls.validate_related_name(model, get_remote_field_model(field), field)
                        # Replace and store the current `on_delete` value to
                        # make sure non-tenant models are not collected on
                        # deletion.
                        on_delete = remote_field.on_delete
                        if on_delete is not DO_NOTHING:
                            remote_field._on_delete = on_delete
                            remote_field.on_delete = DO_NOTHING
                for m2m in opts.local_many_to_many:
                    m2m_remote_field = get_remote_field(m2m)
                    m2m_related_model = get_remote_field_model(m2m)
                    cls.validate_related_name(model, m2m_related_model, m2m)
                    through = m2m_remote_field.through
                    if (not isinstance(through, string_types) and
                            through._meta.auto_created):
                        # Replace the automatically created intermediary model
                        # by a TenantModelBase instance.
                        remove_from_app_cache(through)
                        # Make sure to clear the referenced model cache if
                        # we have contributed to it already.
                        if not isinstance(m2m_related_model, string_types):
                            clear_opts_related_cache(m2m_related_model)
                        m2m_remote_field.through = cls.intermediary_model_factory(m2m, model)
                    else:
                        cls.validate_through(model, m2m_related_model, m2m)
            # Replace `ManagerDescriptor`s with `TenantModelManagerDescriptor`
            # instances.
            for manager in managers:
                setattr(model, manager, TenantModelManagerDescriptor(model))
            # Extract the specified related name if it exists.
            try:
                related_name = attrs.pop('TenantMeta').related_name
            except (KeyError, AttributeError):
                pass
            else:
                # Attach a descriptor to the tenant model to access the
                # underlying model based on the tenant instance.
                def attach_descriptor(tenant_model):
                    descriptor = TenantModelDescriptor(model)
                    setattr(tenant_model, related_name, descriptor)
                app_label, model_name = settings.TENANT_MODEL.split('.')
                lazy_class_prepared(app_label, model_name, attach_descriptor)
            model._for_tenant_model = model
        return model

    @classmethod
    def validate_related_name(cls, model, rel_to, field):
        """
        Make sure that related fields pointing to non-tenant models specify
        a related name containing a %(class)s format placeholder.
        """
        if isinstance(rel_to, string_types):
            lazy_related_operation(cls.validate_related_name, model, rel_to, field=field)
        elif not isinstance(rel_to, TenantModelBase):
            related_name = cls.references[model].related_names[field.name]
            if (related_name is not None and
                    not (get_remote_field(field).is_hidden() or '%(class)s' in related_name)):
                    del cls.references[model]
                    remove_from_app_cache(model, quiet=True)
                    raise ImproperlyConfigured(
                        "Since `%s.%s` is originating from an instance "
                        "of `TenantModelBase` and not pointing to one "
                        "its `related_name` option must ends with a "
                        "'+' or contain the '%%(class)s' format "
                        "placeholder." % (model.__name__, field.name)
                    )

    @classmethod
    def validate_through(cls, model, rel_to, field):
        """
        Make sure the related fields with a specified through points to an
        instance of `TenantModelBase`.
        """
        through = get_remote_field(field).through
        if isinstance(through, string_types):
            lazy_related_operation(cls.validate_through, model, through, field=field)
        elif not isinstance(through, cls):
            del cls.references[model]
            remove_from_app_cache(model, quiet=True)
            raise ImproperlyConfigured(
                "Since `%s.%s` is originating from an instance of "
                "`TenantModelBase` its `through` option must also be pointing "
                "to one." % (model.__name__, field.name)
            )

    @classmethod
    def intermediary_model_factory(cls, field, from_model):
        to_model = get_remote_field_model(field)
        opts = from_model._meta
        from_model_name = opts.model_name
        if to_model == from_model:
            from_ = "from_%s" % from_model_name
            to = "to_%s" % from_model_name
            to_model = from_model
        else:
            from_ = from_model_name
            if isinstance(to_model, string_types):
                to = to_model.split('.')[-1].lower()
            else:
                to = to_model._meta.model_name
        Meta = meta(
            db_table=field._get_m2m_db_table(opts),
            auto_created=from_model,
            app_label=opts.app_label,
            db_tablespace=opts.db_tablespace,
            unique_together=(from_, to),
            verbose_name="%(from)s-%(to)s relationship" % {'from': from_, 'to': to},
            verbose_name_plural="%(from)s-%(to)s relationships" % {'from': from_, 'to': to}
        )
        name = str("%s_%s" % (opts.object_name, field.name))
        field_opts = {'db_tablespace': field.db_tablespace}
        if hasattr(field, 'db_constraint'):
            field_opts['db_constraint'] = field.db_constraint
        return type(name, (cls.tenant_model_class,), {
            'Meta': Meta,
            '__module__': from_model.__module__,
            from_: models.ForeignKey(
                from_model, on_delete=models.CASCADE, related_name="%s+" % name, **field_opts
            ),
            to: models.ForeignKey(
                to_model, on_delete=models.CASCADE, related_name="%s+" % name, **field_opts
            ),
        })

    @classmethod
    def tenant_model_bases(cls, tenant, bases):
        return tuple(
            base.for_tenant(tenant) for base in bases
            if isinstance(base, cls) and not base._meta.abstract
        )

    def abstract_tenant_model_factory(self, tenant):
        if issubclass(self, TenantSpecificModel):
            raise ValueError('Can only be called on non-tenant specific model.')
        reference = self.references[self]
        model = super(TenantModelBase, self).__new__(
            self.__class__,
            str("Abstract%s" % reference.object_name_for_tenant(tenant)),
            (self,) + self.tenant_model_bases(tenant, self.__bases__), {
                '__module__': self.__module__,
                'Meta': meta(
                    reference.Meta,
                    abstract=True
                ),
                tenant.ATTR_NAME: TenantDescriptor(tenant),
                '_for_tenant_model': self
            }
        )
        opts = model._meta

        # Remove ourself from the parents chain and our descriptor
        ptr = opts.parents.pop(self)
        opts.local_fields.remove(ptr)
        delattr(model, ptr.name)

        # Rename parent ptr fields
        for parent, ptr in opts.parents.items():
            local_ptr = self._meta.parents[parent._for_tenant_model]
            ptr.name = None
            ptr.set_attributes_from_name(local_ptr.name)
        # Add copy of the fields to cloak the inherited ones.
        fields = (
            copy.deepcopy(field) for field in (
                self._meta.local_fields +
                self._meta.local_many_to_many +
                get_private_fields(self._meta)
            )
        )
        for field in fields:
            remote_field = get_remote_field(field)
            if remote_field:
                # Make sure related fields pointing to tenant models are
                # pointing to their tenant specific counterpart.
                remote_field_model = get_remote_field_model(field)
                # Clear the field's cache.
                if hasattr(field, '_related_fields'):
                    delattr(field, '_related_fields')
                clear_cached_properties(field)
                clear_cached_properties(remote_field)
                if isinstance(remote_field_model, TenantModelBase):
                    if getattr(remote_field, 'parent_link', False):
                        continue
                    set_remote_field_model(field, self.references[remote_field_model].for_tenant(tenant))
                    # If no `related_name` was specified we make sure to
                    # define one based on the non-tenant specific model name.
                    if not remote_field.related_name:
                        remote_field.related_name = "%s_set" % self._meta.model_name
                else:
                    clear_opts_related_cache(remote_field_model)
                    related_name = reference.related_names[field.name]
                    # The `related_name` was validated earlier to either end
                    # with a '+' sign or to contain %(class)s.
                    if related_name:
                        remote_field.related_name = related_name
                    else:
                        related_name = 'unspecified_for_tenant_model+'
                if isinstance(field, models.ManyToManyField):
                    through = remote_field.through
                    remote_field.through = self.references[through].for_tenant(tenant)
                # Re-assign the correct `on_delete` that was swapped for
                # `DO_NOTHING` to prevent non-tenant model collection.
                on_delete = getattr(remote_field, '_on_delete', None)
                if on_delete:
                    remote_field.on_delete = on_delete
            field.contribute_to_class(model, field.name)

        # Some private fields such as GenericRelation are not correctly
        # cloaked by `contribute_to_class`. Make sure to remove non-tenant
        # private instances from tenant specific model options.
        private_fields = get_private_fields(opts)
        for private_field in get_private_fields(self._meta):
            if private_field in private_fields:
                private_fields.remove(private_field)
            if private_field in opts.local_fields:
                opts.local_fields.remove(private_field)

        return model

    def _prepare(self):
        super(TenantModelBase, self)._prepare()

        if issubclass(self, TenantSpecificModel):
            for_tenant_model = self._for_tenant_model

            # TODO: Remove when dropping support for Django < 1.10
            if django.VERSION >= (1, 10):
                for mgr_name, manager in for_tenant_model._meta.managers_map.items():
                    new_manager = copy.copy(manager)
                    new_manager.creation_counter = manager.creation_counter
                    self.add_to_class(mgr_name, new_manager)
            else:
                for _, mgr_name, manager in for_tenant_model._meta.concrete_managers:
                    new_manager = manager._copy_to_model(self)
                    new_manager.creation_counter = manager.creation_counter
                    self.add_to_class(mgr_name, new_manager)

            # Since our declaration class is not one of our parents we must
            # make sure our exceptions extend his.
            for exception in self.exceptions:
                subclass = subclass_exception(
                    str(exception),
                    (getattr(self, exception), getattr(for_tenant_model, exception)),
                    self.__module__,
                    self,
                )
                self.add_to_class(exception, subclass)

    def for_tenant(self, tenant):
        """
        Returns the model for the specific tenant.
        """
        if issubclass(self, TenantSpecificModel):
            raise ValueError('Can only be called on non-tenant specific model.')
        reference = self.references[self]
        opts = self._meta
        name = reference.object_name_for_tenant(tenant)

        # Return the already cached model instead of creating a new one.
        model = get_model(opts.app_label, name.lower())
        if model:
            return model

        meta_attrs = {
            # TODO: Use `db_schema` once django #6148 is fixed.
            'db_table': db_schema_table(tenant, self._meta.db_table),
            'apps': TenantApps(tenant, getattr(reference.Meta, 'apps', apps)),
        }

        if (1, 10) <= django.VERSION < (2, 0):
            meta_attrs['manager_inheritance_from_future'] = True

        attrs = {
            '__module__': self.__module__,
            'Meta': meta(reference.Meta, **meta_attrs),
        }

        if opts.proxy:
            attrs['_for_tenant_model'] = self

            # In order to make sure the non-tenant model is part of the
            # __mro__ we create an abstract model with stripped fields and
            # inject it as the first base.
            base = type(
                str("Abstract%s" % reference.object_name_for_tenant(tenant)),
                (self,), {
                    '__module__': self.__module__,
                    'Meta': meta(abstract=True),
                }
            )
            # Remove ourself from the parents chain and our descriptor
            base_opts = base._meta
            ptr = base_opts.parents.pop(opts.concrete_model)
            base_opts.local_fields.remove(ptr)
            delattr(base, ptr.name)

            bases = (base,) + self.tenant_model_bases(tenant, self.__bases__)
        else:
            bases = (self.abstract_tenant_model_factory(tenant),)

        model = super(TenantModelBase, self).__new__(
            TenantModelBase, str(name), bases, attrs
        )

        return model

    def destroy(self):
        """
        Remove all reference to this tenant model.
        """
        if not issubclass(self, TenantSpecificModel):
            raise ValueError('Can only be called on tenant specific model.')
        remove_from_app_cache(self, quiet=True)
        if not self._meta.proxy:
            # Some fields (GenericForeignKey, ImageField) attach (pre|post)_init
            # signals to their associated model even if they are abstract.
            # Since this instance was created from an abstract base generated
            # by `abstract_tenant_model_factory` we must make sure to disconnect
            # all signal receivers attached to it in order to be gc'ed.
            disconnect_signals(self.__bases__[0])


def __unpickle_tenant_model_base(model, natural_key, abstract):
    try:
        manager = get_tenant_model()._default_manager
        tenant = manager.get_by_natural_key(*natural_key)
        tenant_model = model.for_tenant(tenant)
        if abstract:
            tenant_model = tenant_model.__bases__[0]
        return tenant_model
    except Exception:
        logger = logging.getLogger('tenancy.pickling')
        logger.exception('Failed to unpickle tenant model')


def __pickle_tenant_model_base(model):
    if issubclass(model, TenantSpecificModel):
        tenant = getattr(model, get_tenant_model().ATTR_NAME)
        return (
            __unpickle_tenant_model_base,
            (model._for_tenant_model, tenant.natural_key(), model._meta.abstract)
        )
    return model.__name__


copyreg.pickle(TenantModelBase, __pickle_tenant_model_base)


class TenantModelDescriptor(object):
    __slots__ = ['model']

    def __init__(self, model):
        self.model = model

    def __get__(self, tenant, owner):
        if not tenant:
            return self
        return tenant.models[self.model]._default_manager


class TenantModel(with_metaclass(TenantModelBase, models.Model)):
    class Meta:
        abstract = True


@receiver(models.signals.class_prepared)
def attach_signals(signal, sender, **kwargs):
    """
    Re-attach signals to tenant models
    """
    if issubclass(sender, TenantSpecificModel):
        for signal, receiver_ in receivers_for_model(sender._for_tenant_model):
            signal.connect(receiver_, sender=sender)


def validate_not_to_tenant_model(model, to, field):
    """
    Make sure the `to` relationship is not pointing to an instance of
    `TenantModelBase`.
    """
    if isinstance(to, string_types):
        lazy_related_operation(validate_not_to_tenant_model, model, to, field=field)
    elif isinstance(to, TenantModelBase):
        remove_from_app_cache(model, quiet=True)
        raise ImproperlyConfigured(
            "`%s.%s`'s `to` option` can't point to an instance of "
            "`TenantModelBase` since it's not one itself." % (
                model.__name__, field.name
            )
        )


@receiver(models.signals.class_prepared)
def validate_relationships(signal, sender, **kwargs):
    """
    Non-tenant models can't have relationships pointing to tenant models.
    """
    if not isinstance(sender, TenantModelBase):
        opts = sender._meta
        # Don't validate auto-intermediary models since they are created
        # before their origin model (from) and cloak the actual, user-defined
        # improper configuration.
        if not opts.auto_created:
            for field in opts.local_fields:
                remote_field = get_remote_field(field)
                if remote_field:
                    validate_not_to_tenant_model(sender, get_remote_field_model(field), field)
            for m2m in opts.local_many_to_many:
                validate_not_to_tenant_model(sender, get_remote_field_model(m2m), m2m)



from __future__ import unicode_literals

from contextlib import contextmanager
from itertools import chain

from django.apps import apps
from django.db import models
from django.utils.functional import cached_property

from .compat import (
    get_deferred_proxies, get_remote_field, get_remote_field_model,
)


def get_model(app_label, model_name):
    try:
        return apps.get_registered_model(app_label, model_name)
    except LookupError:
        pass


@contextmanager
def apps_lock():
    # The registry lock is not re-entrant so we must avoid acquiring it
    # during the initialization phase in order to prevent deadlocks.
    if apps.ready:
        with apps._lock:
            yield
    else:
        yield


def _pop_model_class(model_class, quiet):
    opts = model_class._meta
    apps = opts.apps
    app_label, model_name = opts.app_label, opts.model_name
    try:
        model_class = apps.app_configs[app_label].models.pop(model_name)
    except KeyError:
        if not quiet:
            raise ValueError("%r is not cached" % model_class)
    return model_class


def remove_from_app_cache(model_class, quiet=False):
    with apps_lock():
        model_class = _pop_model_class(model_class, quiet=quiet)
        opts = model_class._meta
        for deferred_proxy in get_deferred_proxies(opts):
            _pop_model_class(deferred_proxy, quiet=True)
        unreference_model(model_class)
        apps.clear_cache()
    return model_class


def get_forward_fields(opts):
    return chain(
        opts.fields,
        opts.many_to_many
    )


def get_reverse_fields(opts):
    return opts._get_fields(forward=False, reverse=True, include_hidden=True)


def clear_opts_related_cache(model_class):
    opts = model_class._meta
    if not opts.apps.ready:
        return
    children = [
        related_object.related_model
        for related_object in opts.__dict__.get('related_objects', []) if related_object.parent_link
    ]
    opts._expire_cache()
    for child in children:
        clear_opts_related_cache(child)


def unreference_model(model):
    disconnect_signals(model)
    for field in get_forward_fields(model._meta):
        if field.model is model:
            remote_field = get_remote_field(field)
            if remote_field:
                remote_field_model = get_remote_field_model(field)
                if isinstance(remote_field_model, models.base.ModelBase):
                    clear_opts_related_cache(remote_field_model)
                    rel_is_hidden = remote_field.is_hidden()
                    # An accessor is added to related classes if they are not
                    # hidden. However o2o fields *always* add an accessor
                    # even if the relationship is hidden.
                    o2o = isinstance(field, models.OneToOneField)
                    if not rel_is_hidden or o2o:
                        try:
                            delattr(remote_field_model, remote_field.get_accessor_name())
                        except AttributeError:
                            pass
            field.model = None


model_sender_signals = (
    models.signals.pre_init,
    models.signals.post_init,
    models.signals.pre_save,
    models.signals.post_save,
    models.signals.pre_delete,
    models.signals.post_delete,
    models.signals.m2m_changed,
)


def receivers_for_model(model):
    for signal in model_sender_signals:
        for receiver in signal._live_receivers(model):
            yield signal, receiver


def disconnect_signals(model):
    for signal, receiver in receivers_for_model(model):
        signal.disconnect(receiver, sender=model)


def clear_cached_properties(instance):
    """
    Clear the cache from the instance properties.
    """
    cls = type(instance)
    for attr in list(instance.__dict__):
        if isinstance(getattr(cls, attr, None), cached_property):
            instance.__dict__.pop(attr)


class SchemaConstraints(object):
    def __init__(self, schema):
        self.schema = schema

    def __call__(self, cursor, table_name):
        """
        Retrieve any constraints or keys (unique, pk, fk, check, index) across
        one or more columns. Also retrieve the definition of expression-based
        indexes.
        """
        constraints = {}
        # Loop over the key table, collecting things as constraints. The column
        # array must return column names in the same order in which they were
        # created.
        # The subquery containing generate_series can be replaced with
        # "WITH ORDINALITY" when support for PostgreSQL 9.3 is dropped.
        cursor.execute("""
            SELECT
                c.conname,
                array(
                    SELECT attname
                    FROM (
                        SELECT unnest(c.conkey) AS colid,
                               generate_series(1, array_length(c.conkey, 1)) AS arridx
                    ) AS cols
                    JOIN pg_attribute AS ca ON cols.colid = ca.attnum
                    WHERE ca.attrelid = c.conrelid
                    ORDER BY cols.arridx
                ),
                c.contype,
                (SELECT fkc.relname || '.' || fka.attname
                FROM pg_attribute AS fka
                JOIN pg_class AS fkc ON fka.attrelid = fkc.oid
                WHERE fka.attrelid = c.confrelid AND fka.attnum = c.confkey[1])
            FROM pg_constraint AS c
            JOIN pg_class AS cl ON c.conrelid = cl.oid
            JOIN pg_namespace AS ns ON cl.relnamespace = ns.oid
            WHERE ns.nspname = %s AND cl.relname = %s
        """, [self.schema, table_name])
        for constraint, columns, kind, used_cols in cursor.fetchall():
            constraints[constraint] = {
                "columns": columns,
                "primary_key": kind == "p",
                "unique": kind in ["p", "u"],
                "foreign_key": tuple(used_cols.split(".", 1)) if kind == "f" else None,
                "check": kind == "c",
                "index": False,
                "definition": None,
            }
        # Now get indexes
        cursor.execute("""
            SELECT
                indexname, array_agg(attname ORDER BY arridx), indisunique, indisprimary,
                array_agg(ordering ORDER BY arridx), amname, exprdef, s2.attoptions
            FROM (
                SELECT
                    c2.relname as indexname, idx.*, attr.attname, am.amname,
                    CASE
                        WHEN idx.indexprs IS NOT NULL THEN
                            pg_get_indexdef(idx.indexrelid)
                    END AS exprdef,
                    CASE am.amname
                        WHEN 'btree' THEN
                            CASE (option & 1)
                                WHEN 1 THEN 'DESC' ELSE 'ASC'
                            END
                    END as ordering,
                    c2.reloptions as attoptions
                FROM (
                    SELECT *
                    FROM pg_index i, unnest(i.indkey, i.indoption) WITH ORDINALITY koi(key, option, arridx)
                ) idx
                LEFT JOIN pg_class c ON idx.indrelid = c.oid
                LEFT JOIN pg_class c2 ON idx.indexrelid = c2.oid
                LEFT JOIN pg_am am ON c2.relam = am.oid
                LEFT JOIN pg_attribute attr ON attr.attrelid = c.oid AND attr.attnum = idx.key
                LEFT JOIN pg_catalog.pg_namespace ns ON c.relnamespace = ns.oid
                WHERE ns.nspname = %s AND c.relname = %s
            ) s2
            GROUP BY indexname, indisunique, indisprimary, amname, exprdef, attoptions;
        """, [self.schema, table_name])
        for index, columns, unique, primary, orders, type_, definition, options in cursor.fetchall():
            if index not in constraints:
                constraints[index] = {
                    "columns": columns if columns != [None] else [],
                    "orders": orders if orders != [None] else [],
                    "primary_key": primary,
                    "unique": unique,
                    "foreign_key": None,
                    "check": False,
                    "index": True,
                    "type": type_,
                    "definition": definition,
                }
        return constraints


@contextmanager
def patch_connection_introspection(connection, schema):
    if connection.vendor == 'postgresql':
        get_constraints = connection.introspection.get_constraints
        connection.introspection.get_constraints = SchemaConstraints(schema)
    try:
        yield
    finally:
        if connection.vendor == 'postgresql':
            connection.introspection.get_constraints = get_constraints



from __future__ import unicode_literals

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

TENANT_MODEL = getattr(settings, 'TENANCY_TENANT_MODEL', 'tenancy.Tenant')

try:
    app_label, model_name = TENANT_MODEL.split('.')
except ValueError:
    raise ImproperlyConfigured(
        "TENANCY_TENANT_MODEL must be of the form 'app_label.model_name'"
    )

HOST_NAME = getattr(settings, 'TENANCY_HOST_NAME', 'tenant')

SCHEMA_AUTHORIZATION = getattr(settings, 'TENANCY_SCHEMA_AUTHORIZATION', False)



from __future__ import unicode_literals

from contextlib import contextmanager
from functools import partial

from django.apps import apps
from django.db.backends.utils import truncate_name
from django.db.migrations import operations
from django.db.migrations.operations.base import Operation
from django.utils import six
from django.utils.six import iteritems

from .models import Managed, db_schema_table
from .utils import patch_connection_introspection


class TenantOperation(Operation):
    def get_tenant_model(self, app_label, from_state, to_state):
        raise NotImplementedError

    @contextmanager
    def tenant_context(self, tenant, schema_editor):
        connection = schema_editor.connection
        cursor = connection.cursor()
        db_schema = tenant.db_schema
        if connection.vendor == 'postgresql':
            sql = "SET search_path = %s, public" % schema_editor.connection.ops.quote_name(tenant.db_schema)
            cursor.execute(sql)
            schema_editor.deferred_sql.append(sql)
        with patch_connection_introspection(connection, db_schema):
            setattr(schema_editor, 'tenant', tenant)
            try:
                yield
            finally:
                delattr(schema_editor, 'tenant')
        if connection.vendor == 'postgresql':
            sql = 'RESET search_path'
            cursor.execute(sql)
            schema_editor.deferred_sql.append(sql)

    def _create_managed_state(self, tenant_model, connection, state):
        managed = Managed("%s.%s" % (tenant_model._meta.app_label, tenant_model._meta.object_name))
        state.__dict__.pop('apps', None)
        project_state = state.clone()
        managed_models = []
        for model_key, model_state in iteritems(project_state.models):
            options = model_state.options
            if options.get('managed') == managed:
                db_table = options.get('db_table')
                if not db_table:
                    db_table = truncate_name("%s_%s" % model_key, connection.ops.max_name_length())
                options.update(
                    managed=True,
                    db_table=db_table,
                )
                managed_models.append(model_key)
        return project_state, managed_models

    def _create_tenant_state(self, tenant, connection, state, managed_models):
        if not managed_models or connection.vendor == 'postgresql':
            return state
        project_state = state.clone()
        for model_key in managed_models:
            model_state = project_state.models[model_key]
            model_state.options['db_table'] = db_schema_table(tenant, model_state.options['db_table'])
        return project_state

    def tenant_operation(self, tenant_model, operation, app_label, schema_editor, from_state, to_state):
        connection = schema_editor.connection
        global_tenant_model = apps.get_model(tenant_model._meta.app_label, tenant_model._meta.model_name)
        get_db_schema = global_tenant_model.db_schema.fget
        get_natural_key = global_tenant_model.natural_key
        if six.PY2:
            get_natural_key = get_natural_key.im_func
        tenants = list(tenant_model._base_manager.all())
        # Small optimization to avoid creating model states if not required.
        if tenants:
            managed_from = self._create_managed_state(tenant_model, connection, from_state)
            managed_to = self._create_managed_state(tenant_model, connection, to_state)
            for tenant in tenants:
                tenant.natural_key = partial(get_natural_key, tenant)
                tenant.db_schema = get_db_schema(tenant)
                tenant_from_state = self._create_tenant_state(tenant, connection, *managed_from)
                tenant_to_state = self._create_tenant_state(tenant, connection, *managed_to)
                with self.tenant_context(tenant, schema_editor):
                    operation(app_label, schema_editor, tenant_from_state, tenant_to_state)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        tenant_model = self.get_tenant_model(app_label, from_state, to_state)
        operation = super(TenantOperation, self).database_forwards
        self.tenant_operation(tenant_model, operation, app_label, schema_editor, from_state, to_state)

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        tenant_model = self.get_tenant_model(app_label, to_state, from_state)
        operation = super(TenantOperation, self).database_backwards
        self.tenant_operation(tenant_model, operation, app_label, schema_editor, from_state, to_state)


class TenantModelOperation(TenantOperation):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return to_state.models[app_label, self.name_lower]

    def get_tenant_model(self, app_label, from_state, to_state):
        model_state = self.get_operation_model_state(app_label, from_state, to_state)
        managed = model_state.options.get('managed')
        return from_state.apps.get_model(managed.tenant_model)


class CreateModel(TenantModelOperation, operations.CreateModel):
    pass


class DeleteModel(TenantModelOperation, operations.DeleteModel):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return from_state.models[app_label, self.name_lower]


class RenameModel(TenantModelOperation, operations.RenameModel):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return from_state.models[app_label, self.old_name_lower]

    database_backwards = operations.RenameModel.database_backwards


class AlterModelTable(TenantModelOperation, operations.AlterModelTable):
    database_backwards = operations.AlterModelTable.database_backwards


class AlterUniqueTogether(TenantModelOperation, operations.AlterUniqueTogether):
    database_backwards = operations.AlterUniqueTogether.database_backwards


class AlterIndexTogether(TenantModelOperation, operations.AlterIndexTogether):
    database_backwards = operations.AlterIndexTogether.database_backwards


class TenantModelFieldOperation(TenantModelOperation):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return from_state.models[app_label, self.model_name_lower]


class AddField(TenantModelFieldOperation, operations.AddField):
    pass


class RemoveField(TenantModelFieldOperation, operations.RemoveField):
    pass


class AlterField(TenantModelFieldOperation, operations.AlterField):
    database_backwards = operations.AlterField.database_backwards


class RenameField(TenantModelFieldOperation, operations.RenameField):
    pass


class TenantSpecialOperation(TenantOperation):
    def __init__(self, tenant_model, *args, **kwargs):
        self.tenant_model = tenant_model
        super(TenantSpecialOperation, self).__init__(*args, **kwargs)

    def get_tenant_model(self, app_label, from_state, to_state):
        opts = self.tenant_model._meta
        return from_state.apps.get_model(opts.app_label, opts.model_name)


class RunPython(TenantSpecialOperation, operations.RunPython):
    pass


class RunSQL(TenantSpecialOperation, operations.RunSQL):
    pass



from __future__ import unicode_literals

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import DEFAULT_DB_ALIAS, connections
from django.http import Http404

from . import get_tenant_model
from .settings import HOST_NAME

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object


class TenantHostMiddleware(MiddlewareMixin):
    def __init__(self, *args, **kwargs):
        super(TenantHostMiddleware, self).__init__(*args, **kwargs)
        try:
            import django_hosts  # NOQA
        except ImportError:
            raise ImproperlyConfigured(
                'You must install django-hosts in order to use `TenantHostMiddleware`.'
            )
        django_hosts_middleware = 'django_hosts.middleware.HostsRequestMiddleware'
        path = "%s.%s" % (self.__module__, self.__class__.__name__)
        for setting in ('MIDDLEWARE', 'MIDDLEWARE_CLASSES'):
            middlewares = getattr(settings, setting, None)
            if middlewares is None:
                continue
            for middleware in middlewares:
                if middleware == django_hosts_middleware:
                    break
                elif middleware == path:
                    raise ImproperlyConfigured(
                        "Make sure '%s' appears before '%s' in your `%s` setting." % (
                            django_hosts_middleware, path, setting,
                        )
                    )
            break
        self.tenant_model = get_tenant_model()
        self.attr_name = self.tenant_model.ATTR_NAME

    def process_request(self, request):
        if request.host.name == HOST_NAME:
            match = request.host.compiled_regex.match(request.get_host())
            lookups = match.groupdict()
            tenant_model = self.tenant_model
            try:
                tenant = tenant_model._default_manager.get(**lookups)
            except tenant_model.DoesNotExist:
                raise Http404(
                    "No tenant found for specified lookups: %r" % lookups
                )
            setattr(request, self.attr_name, tenant)


class GlobalTenantMiddleware(MiddlewareMixin):
    """
    Middleware that assigns the request's tenant attribute to the default
    connection object. This unfortunate global state is required in order to
    allow things such as a tenant custom user with the required auth backend.
    """

    def __init__(self, *args, **kwargs):
        super(GlobalTenantMiddleware, self).__init__(*args, **kwargs)
        self.attr_name = get_tenant_model().ATTR_NAME

    def get_global_state(self):
        return connections[DEFAULT_DB_ALIAS]

    def pollute_global_state(self, tenant):
        setattr(self.get_global_state(), self.attr_name, tenant)

    def clean_global_state(self):
        global_state = self.get_global_state()
        if hasattr(global_state, self.attr_name):
            delattr(global_state, self.attr_name)

    def process_request(self, request):
        self.pollute_global_state(
            getattr(request, self.attr_name, None)
        )

    def process_response(self, request, response):
        self.clean_global_state()
        return response

    def process_exception(self, request, exception):
        self.clean_global_state()
